{-# LANGUAGE DataKinds #-}

module Fedi.Requests where

import Control.Logger.Simple qualified as Log
import Data.Aeson qualified as A
import Data.ByteString.Lazy qualified as BSL
import Data.List (intercalate)
import Data.Text qualified as T
import Fedi.Helpers
import Fedi.Signature.Sign
import Fedi.UserDetails
import Network.HTTP.Req qualified as Req
import Text.URI qualified as URI

sendPost
  :: (A.ToJSON input)
  => UserDetails
  -> String
  -> input
  -> IO ByteString
sendPost details url payload = do
  uri <- URI.mkURI $ fromString url
  Log.logDebug $ "Post To: " <> fromString url
  Log.logDebug $ "Post Sending: " <> pJson payload

  let
    encoded = BSL.toStrict $ A.encode payload
  httpSignature <- makeHttpSignature details uri encoded
  Log.logDebug $ "Post http signature: " <> pShow httpSignature
  Log.logDebug $ "Post http signature headers: " <> pShow (makeSigHeaders httpSignature)
  (url', scheme) <- maybe (throw "couldn't parse uri") pure (Req.useHttpsURI uri)
  Req.runReq Req.defaultHttpConfig do
    response <-
      Req.reqCb
        Req.POST
        url'
        (Req.ReqBodyBs encoded)
        Req.bsResponse
        ( scheme
            <> sigHeaders httpSignature
        )
        ( \request -> do
            Log.logDebug $ "Sending POST request: " <> pShow request
            pure request
        )
    Log.logInfo $ "Post Response: " <> pShow response
    pure $ Req.responseBody response

makeHttpSignature :: UserDetails -> URI.URI -> ByteString -> IO HttpSignature
makeHttpSignature details uri encoded = do
  host <- case uri.uriAuthority of
    Right u -> pure $ T.unpack $ URI.unRText u.authHost
    _ -> throw "no host in uri"
  -- We may want to add the uri path and uri query,
  -- but maybe later.
  let
    path = case uri.uriPath of
      Just (_, list) ->
        "/" <> intercalate "/" (map (T.unpack . URI.unRText) $ toList list)
      Nothing -> "/"
  signSignature details host ("post " <> path) encoded

sigHeaders :: HttpSignature -> Req.Option scheme
sigHeaders =
  foldMap (uncurry Req.header) . makeSigHeaders

makeSigHeaders :: HttpSignature -> [(ByteString, ByteString)]
makeSigHeaders httpSignature =
  [ ("ContentType", "application/activity+json")
  , ("Accept", "application/activity+json")
  , ("Digest", httpSignature.digest)
  , ("Host", fromString httpSignature.host)
  , ("Date", fromString httpSignature.date)
  , ("Signature", toSignature httpSignature.signatureHeader)
  ]

sendGet :: (Show a, A.FromJSON a) => String -> IO a
sendGet url = do
  uri <- URI.mkURI $ fromString url
  (url', scheme) <- maybe (throw "couldn't parse uri") pure (Req.useHttpsURI uri)

  Req.runReq Req.defaultHttpConfig do
    response <-
      Req.reqCb
        Req.GET
        url'
        Req.NoReqBody
        Req.jsonResponse
        ( scheme
            <> Req.header "ContentType" "application/activity+json"
        )
        ( \request -> do
            Log.logDebug $ "Sending GET request: " <> pShow request
            pure request
        )
    Log.logInfo $ "Get Response: " <> pShow response
    pure $ Req.responseBody response
